/**
  *  \file client/widgets/processlist.cc
  *  \brief Process List Widget
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/processlist.h"
#include "gfx/gfxcontext.h"
#include "int/process.h"
#include "int/exec.h"
#include "util/nls.h"
#include "util/misc.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "int/notify.h"

namespace {
    enum {
        NameWidth   = 20,
        MsgWidth    = 3,
        WhereWidth  = 5,
        PriWidth    = 2,
        StatusWidth = 8,

        TotalWidth = NameWidth + MsgWidth + WhereWidth + PriWidth + StatusWidth
    };
}

static const char*const stateNameTable[] = {
    N_("Running"),
    N_("Runnable"),
    N_("Terminated"),
    N_("Ended"),
    N_("Suspended"),
    N_("Error"),
    N_("Frozen"),
    N_("Terminated"),
};

/****************************** WProcessList *****************************/

/** Constructor. */
WProcessList::WProcessList()
    : UIListbox(0, 14 /* same as PCC 1.x */, TotalWidth*font_ems[FONT_NORMAL], 0)
{ }

WProcessList::~WProcessList()
{ }

void
WProcessList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    const IntProcessList& pl = getProcessList();
    const int em = font_ems[FONT_NORMAL];
    for (int i = from; i < to; ++i) {
        // Prepare line
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        // Do we have a process here?
        int slot = getTop() + i;
        if (slot >= 0 && slot < int(pl.size())) {
            // Set color: frozen tasks are faded because they cannot be manipulated (but examined)
            const IntExecutionContext& process = *pl[slot];
            if (process.getState() == IntExecutionContext::psFrozen) {
                ctx.setColor(UIColor::tc_Faded);
            }

            // Figure out location
            GObject* o = process.getInvokingObject();
            string_t where;
            if (dynamic_cast<GPlanet*>(o)) {
                if (process.getProcessKind() == IntExecutionContext::pkBaseTask) {
                    where = format("b%d", o->getId());
                } else {
                    where = format("p%d", o->getId());
                }
            } else if (dynamic_cast<GShip*>(o)) {
                where = format("s%d", o->getId());
            }

            // Message status
            string_t msg;
            int msgfont = FONT_NORMAL;
            if (IntNotificationMessage* m = process.getNotificationMessage()) {
                if (m->isConfirmed()) {
                    msg = _("OK");
                } else {
                    msg = _("New");
                    msgfont = FONT_NORMAL_BOLD;
                }
            }

            // Show it
            area.consumeX(5);
            outTextF(ctx, area.splitX(NameWidth*em), process.getName());
            ctx.useFont(*fonts[msgfont]);
            outTextF(ctx, area.splitX(MsgWidth*em), msg);
            ctx.useFont(*fonts[FONT_NORMAL]);
            outTextF(ctx, area.splitX(WhereWidth*em), where);
            outTextF(ctx, area.splitX(PriWidth*em), numToString(process.getPriority()));
            outTextF(ctx, area, _(stateNameTable[process.getState()]));
        }
    }
}

void
WProcessList::update()
{
    setSize(getProcessList().size());
    adjustPosition();
    drawWidget();
    sig_moved.raise(getId());
}

void
WProcessList::scrollToProcess(IntExecutionContext* proc)
{
    const IntProcessList& pl = getProcessList();
    for (uint32_t i = 0; i < pl.size(); ++i) {
        if (pl[i] == proc) {
            scrollTo(i, 1);
            break;
        }
    }
}

IntExecutionContext*
WProcessList::getSelectedProcess()
{
    const IntProcessList& pl = getProcessList();
    if (getCurrent() < int(pl.size())) {
        return pl[getCurrent()];
    } else {
        return 0;
    }
}

/*************************** WProcessListHeader **************************/

WProcessListHeader::WProcessListHeader()
{ }

WProcessListHeader::~WProcessListHeader()
{ }

void
WProcessListHeader::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    const int em = font_ems[FONT_NORMAL];
    int x = getExtent().x;
    int y = getExtent().y;
    outTextF(ctx, x, y, NameWidth*em, _("Name"));
    x += NameWidth*em + 5;
    outTextF(ctx, x, y, MsgWidth*em, _("Msg"));
    x += MsgWidth*em;
    outTextF(ctx, x, y, WhereWidth*em, _("Where"));
    x += WhereWidth*em;
    outTextF(ctx, x, y, PriWidth*em, _("Pri"));
    x += PriWidth*em;
    outTextF(ctx, x, y, StatusWidth*em, _("Status"));
}

void
WProcessListHeader::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(TotalWidth, 1);
    info.flags = 0;
}
