/**
  *  \file game/obj.h
  *  \brief Game Object Base Classes
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJ_H
#define PCC_V2_GAME_OBJ_H

#include "game/coord.h"
#include "cpluslib/string.h"
#include "cpluslib/object.h"
#include "cpluslib/types.h"
#include "cpluslib/sigslot.h"

/** Game object base class.
    A game object has the following basic properties:
    - a name (coming in two flavours, regular and full)
    - an owner
    - an existance flag

    Most code deals with existant objects only.
    However, we allow non-existant GObject descendants to track history.

    Each object can have the following flags:
    - played: if set, the object is "played" by the user, i.e. belongs to him.
      This implies that the object exists.
    - editable: if set, the object can be edited. This applies to all played
      objects from the current turn. Played objects from previous/future turns
      are not editable. Editability implies playability.
    - marked: this is the user-visible object selection

    Finally, objects have a "dirty" flag used to track changes,
    and a sig_changed signal to allow others to hook into these changes.
    Actual change notification is done by GUniverse. */
class GObject : public Object {
 public:
    /** A name. Identifies a name that can be queried with getName(). */
    enum Name {
        PlainName,                 /**< Just the unit's name. Example: "USS Dull" */
        LongName,                  /**< Long name, including the Id number. Example: "Ship #1: USS Dull" */
        DetailedName               /**< Name with details. Example: "Ship #1: USS Dull (Fed Outrider)" or "Ship #1: USS Dull: colonizer" */
    };

    /** Playability level. */
    enum Playability {
        NotPlayable,               /**< Not playable. For example a foreign ship. */
        ReadOnly,                  /**< Read only. Used to be playable, just not now. For example, our ship in a past turn. */
        Playable,                  /**< Playable. Can be manipulated according to rules. For example, our ship. */
        Editable                   /**< Editable. Can be manipulated, also outside the rules. */
    };

    /** Constructor. */
    GObject()
        : playable(NotPlayable), is_marked(false), is_dirty(false)
        { }

    /** Destructor. */
    virtual ~GObject();

    // Object:
    /** Convert to string. */
    string_t toString();

    // GObject interface:
    /** Get name of this object.
        \param which Which name format to return */
    virtual string_t getName(Name which) const = 0;

    /** Get Id number of this object. */
    virtual int getId() const = 0;

    /** Get owner of this object. */
    virtual int getOwner() const = 0;


    /** Check whether this object is being played.
        -- old interface, deprecated -- */
    bool isPlayed() const
        { return playable >= Playable; }

    /** Check playability level. */
    bool isPlayable(Playability p) const
        { return playable >= p; }

    /** Set playability. */
    void setPlayability(Playability p)
        { playable = p; }

    /** Get playability. */
    Playability getPlayability() const
        { return playable; }

    /** Mark object clean. */
    void markClean()
        { is_dirty = false; }

    /** Mark object dirty. */
    void markDirty()
        { is_dirty = true; }

    /** Check whether specific area of object is dirty. */
    bool isDirty() const
        { return is_dirty; }

    
    /** Check whether object is marked. */
    bool isMarked() const
        { return is_marked; }

    /** Set selection status. */
    void setIsMarked(bool n)
        {
            markDirty();
            is_marked = n;
        }

    void notifyListeners();

    /** Signal for object changes.
        \param int getId() of the object */
    Signal1<int> sig_changed;

 private:
    Playability playable : 8;       /**< Playability. */
    bool is_marked;                 /**< User tag. */
    bool is_dirty;                  /**< Dirty flag. */

    GObject(const GObject&);
    GObject& operator=(const GObject&);
};


/** Game object on the map.
    This extends GObject to include a map position. */
class GMapObject : public GObject {
 public:
    /** Destructor. */
    virtual ~GMapObject();

    /** Get position in game universe. */
    virtual GPoint getPos() const = 0;
};


/** Circular game object on the map.
    This extends GObject to include a radius. */
class GCircularObject : public GMapObject {
 public:
    virtual ~GCircularObject();

    /** Get object radius in light-years. */
    virtual int getRadius() const = 0;

    /** Get squared radius in light-years. This is used for objects that have fractional radius,
        such as minefields. */
    virtual int32_t getRadiusSquared() const = 0;
};

#endif
