/**
  *  \file game/ufo.h
  *  \brief Ufo and Wormhole handling
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UFO_H
#define PCC_V2_GAME_UFO_H

#include <list>
#include "game/obj.h"
#include "cpluslib/string.h"
#include "game/objtypes.h"
#include "game/utildata.h"

struct TUfo;
struct TDbUfo;
struct TUtil14Wormhole;
struct TUtil33GO;

/** Ufo. Represents Winplan Ufos, PHost General Objects, and PHost Wormholes. */
class GUfo : public GCircularObject {
 public:
    GUfo(int id);
    GUfo(const GUfo& other);
    ~GUfo();

    GUfo& operator=(const GUfo& rhs);

    /* GCircularObject interface */
    string_t getName(Name which) const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;
    int      getRadius() const;
    int32_t  getRadiusSquared() const;

    /* Loader interface */
    void addUfoData(const TUfo& data);
    void addObjectData(const TUtil33GO& obj);
    void addHistoryData(const TDbUfo& data);
    void addWormholeData(const TUtil14Wormhole& data, bool all, int turn);

    void disconnect();
    void connectWith(GUfo& other);
    void postprocess(int turn);

    bool isStoredInHistory() const;
    void getHistoryData(TDbUfo& data) const;
    void setIsStoredInHistory(bool value);

    bool isSeenThisTurn() const;

    /* Accessors */
    int      getColorCode() const;
    uint8    getColor() const;
    int      getSpeed() const;
    int      getHeading() const;
    int      getPlanetRange() const;
    int      getShipRange() const;
    int      getTypeCode() const;
    string_t getInfo1() const;
    string_t getInfo2() const;
    int32    getRealId() const;
    GPoint   getOtherEndPos() const;
    int      getLastTurn() const;
    GPoint   getMovementVector() const;
    GUfo*    getOtherEnd() const;

    void     setRealId(int32_t real_id);
    void     setMovementVector(GPoint movement_vector);

    /* Values for flags */
    enum {
        UfoSeenThisTurn     = 1,
        UfoStoredInDatabase = 2
    };

 private:
    /* Standard Ufo properties */
    int16  id;
    int16  vga_color;
    GPoint pos;
    int16  warp;
    int16  heading;
    int16  planet_range, ship_range;
    int16  radius;
    int16  type_code;
    string_t name;
    string_t info1, info2;

    /* Additional PCC properties */
    int32  real_id;
    int16  turn_last_seen;
    GPoint pos_last_seen;
    GPoint movement_vector;

    int    flags;

    GUfo* other_end;

    void copyUfo(const TUfo& data);
};

/** Ufo type. Acts as a GObjectType and container for GUfo's. */
class GUfoType : public GObjectType {
 public:
    // Types:
    typedef std::list<GUfo> ufolist_t;
    typedef std::list<TUtil14Wormhole> wormlist_t;
    typedef ufolist_t::iterator iterator;
    typedef ufolist_t::const_iterator const_iterator;

    // Construction:
    GUfoType(GUniverse& univ);
    ~GUfoType();

    // Loading:
    void addWormholeData(const TUtil14Wormhole& wh);
    void addUfoData(int id, const TUfo& ufo);
    void addObjectData(const TUtil33GO& obj);
    void addHistoryData(const TDbUfo& ufo);
    void postprocess();

    // GObjectType:
    GUfo& getObjectByIndex(int index);
    GUniverse* getUniverseByIndex(int index);
    bool isValidIndex(int index) const;
    int getNextIndex(int index) const;
    int getPreviousIndex(int index) const;

    // Container access:
    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;

    iterator getFirstAt(int id);

 private:
    GUfo& getUfoById(int id);
    GUfo& getUfoByIndex(int32_t index);

    // Universe link
    GUniverse& univ;

    // Ufo storage
    ufolist_t ufos;
    int32_t num_ufos;

    // Wormhole storage
    wormlist_t wormholes;
};

inline bool
GUfo::isStoredInHistory() const
{
    return (flags & UfoStoredInDatabase) != 0;
}

inline bool
GUfo::isSeenThisTurn() const
{
    return (flags & UfoSeenThisTurn) != 0;
}

inline GUfoType::iterator
GUfoType::begin()
{
    return ufos.begin();
}

inline GUfoType::iterator
GUfoType::end()
{
    return ufos.end();
}

inline GUfoType::const_iterator
GUfoType::begin() const
{
    return ufos.begin();
}

inline GUfoType::const_iterator
GUfoType::end() const
{
    return ufos.end();
}

#endif
