/**
  *  \file reshack/fontdlg.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "reshack/fontdlg.h"
#include "reshack/charname.h"
#include "gfx/gfxcontext.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "cpluslib/ptrvector.h"
#include "util/recode.h"
#include "gfx/bmfont.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/msgbox.h"
#include "cpluslib/utf8.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/multilinestatic.h"
#include "util/misc.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"

namespace {
    class RHFontConverageWidget : public UIWidget {
     public:
        RHFontConverageWidget(string_t encodingName,
                              uint32_t numMissing,
                              uint32_t firstMissing);

        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
        bool handleEvent(const UIEvent& event, bool second_pass);

        uint32_t getNumMissing() const { return numMissing; }
        uint32_t getFirstMissing() const { return firstMissing; }
     private:
        const string_t encodingName;
        const uint32_t numMissing;
        const uint32_t firstMissing;
    };

    class RHFontAlignWindow : public UIWindow {
        UIInputLine* input_top;
        UIInputLine* input_bottom;
        ObjectHolder h;
        int result_top;
        int result_bottom;
     public:
        RHFontAlignWindow(int top, int bot);
        void init(int top, int bot);
        void onOK();
        int getResultTop() const
            { return result_top; }
        int getResultBottom() const
            { return result_bottom; }
    };
}

static const uint16_t pcc_chars[] = {
    0x00AE, 1,                  // REGISTERED SIGN
    0x00B1, 1,                  // PLUS MINUS SIGN
    0x00B7, 1,                  // MIDDLE DOT
    0x00D7, 1,                  // MULTIPLICATION SIGN
    0x2013, 1,                  // EN DASH
    0x2022, 2,                  // BULLET (round), TRIANGULAR BULLET
    0x2122, 1,                  // TRADEMARK SIGN
    0x2190, 6,                  // LEFTWARDS ARROW, UPWARDS ARROW, RIGHTWARDS ARROW, DOWNWARDS ARROW, LEFT RIGHT ARROW, UP DOWN ARROW
    0x219E, 4,                  // LEFTWARDS TWO HEADED ARROW, UPWARDS TWO HEADED ARROW, RIGHTWARDS TWO HEADED ARROW, DOWNWARDS TWO HEADED ARROW
    0x2245, 1,                  // APPROXIMATELY EQUAL TO (=~)
    0x2248, 1,                  // ALMOST EQUAL TO (~~)
    0x2259, 1,                  // ESTIMATES (corresponds to) (=^)
    0x2264, 2,                  // LESS-THAN OR EQUAL TO (-<), GREATER-THAN OR EQUAL TO (->)
    0x25A0, 1,                  // SQUARE BULLET
    0x25B6, 1,                  // BLACK RIGHT-POINTING TRIANGLE (play)
    0x25BA, 1,                  // BLACK RIGHT-POINTING POINTER (bullet)
    0x25C0, 1,                  // BLACK LEFT-POINTING TRIANGLE (rew)
    0x2713, 2,                  // CHECK MARK
    0x2717, 1,                  // BALLOT CROSS
    0xE100, 16,                 // [PCC2] REPLACEMENT UPPER LEFT 0 (we only require the upper-left glyphs)
    0xE140, 5,                  // [PCC2] TOP ARC CLOCKWISE ARROW, NESTED RECTANGLES ORNAMENT, ORNAMENT LEFT, ORNAMENT RIGHT, THIN VERTICAL BAR (goes with BLACK TRIANGLE to form rewind-to-beginning)
    0,
};
static const uint16_t wgl4_chars[] = {
    0x0020, 0x5F,               // Basic latin
    0x00A0, 0x60,               // Latin-1 supplement
    0x0100, 0x80,               // Latin Extended A
    0x0192, 1,                  // Latin Extended B
    0x01FA, 6,                  // Latin Extended B
    0x02C6, 2,                  // Spacing Modifier Letters
    0x02C9, 1,
    0x02D8, 6,
    0x0384, 7,                  // Greek
    0x038C, 1,
    0x038E, 20,
    0x03A3, 44,
    0x0400, 0x60,               // Cyrillic
    0x0490, 2,
    0x1E80, 6,                  // Latin Extended Additional
    0x1EF2, 2,
    0x2013, 3,                  // General Punctuation
    0x2017, 8,
    0x2020, 3,
    0x2026, 1,
    0x2030, 1,
    0x2032, 2,
    0x2039, 2,
    0x203C, 1,
    0x203E, 1,
    0x2044, 1,
    0x207F, 1,                  // Superscripts and Subscripts
    0x20A3, 2,                  // Currency Symbols
    0x20A7, 1,
    0x20AC, 1,
    0x2105, 1,                  // Letterlike Symbols
    0x2113, 1,
    0x2116, 1,
    0x2122, 1,
    0x2126, 1,
    0x212E, 1,
    0x215B, 4,                  // Number Forms
    0x2190, 6,                  // Arrows
    0x21A8, 0,
    0x2202, 1,                  // Mathematical Operators
    0x2206, 1,
    0x220F, 1,
    0x2211, 2,
    0x2215, 1,
    0x2219, 2,
    0x221F, 2,
    0x2229, 1,
    0x222B, 1,
    0x2248, 1,
    0x2260, 2,
    0x2264, 2,
    0x2302, 1,                  // Miscellaneous Technical
    0x2310, 1,
    0x2320, 2,
    0x2500, 1,                  // Box drawing
    0x2502, 1,
    0x250C, 1,
    0x2510, 1,
    0x2514, 1,
    0x2518, 1,
    0x251C, 1,
    0x2524, 1,
    0x252C, 1,
    0x2534, 1,
    0x253C, 1,
    0x2550, 0x1D,
    0x2580, 1,                  // Block Elements
    0x2584, 1,
    0x2588, 1,
    0x258C, 1,
    0x2590, 4,
    0x25A0, 2,                  // Geometric Shapes
    0x25AA, 3,
    0x25B2, 1,
    0x25BA, 1,
    0x25C4, 1,
    0x25CA, 2,
    0x25CF, 1,
    0x25D8, 2,
    0x25E6, 1,
    0x263A, 3,                  // Miscellaneous Symbols
    0x2640, 1,
    0x2642, 1,
    0x2660, 1,
    0x2663, 1,
    0x2665, 2,
    0x266A, 2,
    // 0xF001, 2,               // Private Use Area
    0xFB01, 2,                  // Alphabetic Presentation Forms (fi, fl)
    0,
};

static const uint16_t*const extra_charsets[] = {
    pcc_chars,
    wgl4_chars,
};
static const char*const extra_charset_names[] = {
    "PCC2 Extras",
    "WGL4",
};

RHFontConverageWidget::RHFontConverageWidget(string_t encodingName,
                                             uint32_t numMissing,
                                             uint32_t firstMissing)
    : encodingName(encodingName),
      numMissing(numMissing),
      firstMissing(firstMissing)
{ }

void
RHFontConverageWidget::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, extent.x, extent.y, extent.w/2, format(_("%s: %d missing"), encodingName, numMissing));

    if (numMissing == 0) {
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, extent.x+extent.w/2, extent.y, extent.w/2, _("complete"));
    } else {
        ctx.setColor(UIColor::tc_Red);
        outTextF(ctx, extent.x+extent.w/2, extent.y, extent.w/2, getCharacterName(firstMissing));
    }

    can.update(getExtent());
}

void
RHFontConverageWidget::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(30, 1);
}

bool
RHFontConverageWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    return UIWidget::handleEvent(event, second_pass);
}


bool
doFontCoverageDialog(const GfxBitmapFont& font, uint32_t& result)
{
    static int lastFocus = 0;

    /* Figure out information */
    ptr_vector<RHFontConverageWidget> info;

    // First: ASCII
    uint32_t numMissing = 0, firstMissing = 0;
    for (int i = 32; i < 127; ++i) {
        if (font.getGlyph(i) == 0) {
            ++numMissing;
            if (!firstMissing)
                firstMissing = i;
        }
    }
    info.push_back(new RHFontConverageWidget("ASCII", numMissing, firstMissing));

    // Now the character sets
    for (std::size_t x = 0; x < CharacterSet::getNumCharsets(); ++x) {
        CharacterSet cs = CharacterSet::getCharset(x);
        numMissing = firstMissing = 0;
        for (int i = 128; i < 256; ++i) {
            uint16_t ch = cs.getUnicode(i);
            if (ch >= 0xA0 && font.getGlyph(ch) == 0) {
                ++numMissing;
                if (!firstMissing)
                    firstMissing = ch;
            }
        }
        info.push_back(new RHFontConverageWidget(cs.getName(), numMissing, firstMissing));
    }

    // Finally, our specials
    for (std::size_t x = 0; x < countof(extra_charsets); ++x) {
        numMissing = firstMissing = 0;
        const uint16_t* p = extra_charsets[x];
        while (*p != 0) {
            uint16_t ch = *p++;
            uint16_t count = *p++;
            while (count != 0) {
                if (font.getGlyph(ch) == 0) {
                    ++numMissing;
                    if (!firstMissing)
                        firstMissing = ch;
                }
                ++ch, --count;
            }
        }
        info.push_back(new RHFontConverageWidget(extra_charset_names[x], numMissing, firstMissing));
    }

    /* Build the window */
    ObjectHolder h;

    // VBox
    //   UIFocusGroup / RHFontConverageWidget *
    //   HBox
    //     UISpacer
    //     "OK"
    //     "Cancel"
    UIWindow& win = h.add(new UIWindow(_("Character Coverage"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);
    for (std::size_t i = 0; i < info.size(); ++i) {
        UIFocusGroup& fg = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 1000+i));
        fg.add(*info[i]);
        win.add(fg);
    }

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    win.add(g);
    win.add(h.add(new UIFocusIterator(1000, 1000 + info.size() - 1, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Home | UIFocusIterator::fi_Wrap)));
    if (UIBaseWidget* b = win.getWidgetById(lastFocus + 1000))
        b->setFocusedWidget();

    win.pack();
    ui_root->centerWidget(win);
    bool ok = ui_root->runModal(win);

    if (ok
        && win.getFocusedChild() != 0
        && win.getFocusedChild()->getId() >= 1000
        && win.getFocusedChild()->getId() < int(1000 + info.size())
        && info[win.getFocusedChild()->getId() - 1000]->getNumMissing() != 0)
    {
        lastFocus = win.getFocusedChild()->getId() - 1000;
        result = info[lastFocus]->getFirstMissing();
        return true;
    } else {
        return false;
    }
}

void
doCharacterCodeInfo(uint32_t code)
{
    string_t text;
    text += format(_("Unicode: U+%04X (%0$0d)\n"), code);
    if (code < 128) {
        // ASCII
        text += format(_("This is an ASCII character.\n"
                         "C encoding: '\\%03o', '\\x%0$02X'\n"), code);
    } else {
        // Unicode
        string_t tmp;
        appendUtf8(tmp, code);
        text += _("In UTF-8: '");
        for (string_t::size_type i = 0; i < tmp.size(); ++i)
            text += format("\\%03o", uint8_t(tmp[i]));
        text += "', '";
        for (string_t::size_type i = 0; i < tmp.size(); ++i)
            text += format("\\x%02X", uint8_t(tmp[i]));
        text += "'\n";

        // Codepages
        bool did = false;
        for (std::size_t x = 0; x < CharacterSet::getNumCharsets(); ++x) {
            CharacterSet cs = CharacterSet::getCharset(x);
            char c = cs.getCharacter(code);
            if (c != 0) {
                text += format(_("In %s: %d ('\\%1$03o', '\\x%1$02X')\n"), cs.getName(), uint8_t(c));
                did = true;
            }
        }
        if (!did) {
            text += _("Not in any codepage.\n");
        }
    }
    messageBox(text, _("Encoding Information"));
}

RHFontAlignWindow::RHFontAlignWindow(int top, int bot)
    : UIWindow(_("Vertical Alignment"), BLUE_WINDOW, COLOR_GRAY)
{
    init(top, bot);
}

void
RHFontAlignWindow::init(int top, int bot)
{
    // VBox
    //   Grid
    //     "Top:"    UIInputLine
    //     "Bottom:" UIInputLine
    //   UIMultilineStatic
    //   HBox
    //     UISpacer
    //     "OK"
    //     "Cancel"
    UIGroup& g1 = h.add(new UIGroup(h.add(new UIGridLayout(2))));
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    input_top    = &h.add(new UIInputLine(10, ss_Alt + 't', 0, 1000));
    input_bottom = &h.add(new UIInputLine(10, ss_Alt + 'b', 0, 1001));
    g1.add(h.add(new UIStaticText(_("Top:"), UIStaticText::StaticTitle)));
    g1.add(*input_top);
    g1.add(h.add(new UIStaticText(_("Bottom:"), UIStaticText::StaticTitle)));
    g1.add(*input_bottom);

    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &RHFontAlignWindow::onOK));
    g3.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(g1);
    add(h.add(new UIMultilineStatic(FONT_NORMAL,
                                    format(_("Enter number of pixels to add at top/bottom. "
                                             "Negative numbers can be used to remove pixels. "
                                             "Maximum you can remove without clipping from the top is "
                                             "%d pixels on the top, %d pixels at the bottom."),
                                           top, bot),
                                    25*font_ems[FONT_NORMAL])));
    add(g3);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Vert)));
    setLayout(UIVBoxLayout::instance5);
    pack();
}

void
RHFontAlignWindow::onOK()
{
    if (!stringValue(input_top->getText(), result_top)) {
        input_top->setFocusedWidget();
        return;
    }
    if (!stringValue(input_bottom->getText(), result_bottom)) {
        input_bottom->setFocusedWidget();
        return;
    }
    stop(1);
}

Ptr<GfxBitmapFont>
doVerticalAlignDialog(const GfxBitmapFont& font)
{
    // Figure out limits
    int top = font.getHeight();
    int bot = 0;
    for (uint32_t i = 0, e = font.getCurrentCharacterLimit(); i < e; ++i) {
        if (const GfxBitmapGlyph* g = font.getGlyph(i)) {
            for (int y = 0; y < g->getHeight(); ++y) {
                for (int x = 0; x < g->getWidth(); ++x) {
                    if (g->get(x, y)) {
                        if (y < top)
                            top = y;
                        if (y > bot)
                            bot = y+1;
                    }
                }
            }
        }
    }

    // Dialog
    RHFontAlignWindow win(top, font.getHeight()-bot);
    ui_root->centerWidget(win);
    if (!ui_root->runModal(win))
        return 0;

    // Compute new height
    int newTop = win.getResultTop();
    int newBottom = win.getResultBottom();
    int newHeight = font.getHeight() + newTop + newBottom;
    if (newHeight <= 0) {
        messageBox(_("The values you entered make the new font have a height of 0 or less."), _("Vertical Alignment"));
        return 0;
    }

    // Do it
    Ptr<GfxBitmapFont> nf = new GfxBitmapFont();
    for (uint32_t i = 0, e = font.getCurrentCharacterLimit(); i < e; ++i) {
        if (const GfxBitmapGlyph* g = font.getGlyph(i)) {
            // Construct new glyph
            std::auto_ptr<GfxBitmapGlyph> ng(new GfxBitmapGlyph(g->getWidth(), newHeight));
            for (int y = 0; y < newHeight; ++y) {
                if (y - newTop >= 0 && y - newTop < g->getHeight()) {
                    for (int x = 0; x < g->getWidth(); ++x) {
                        if (g->get(x, y - newTop))
                            ng->set(x, y, true);
                    }
                }
            }
            const std::vector<uint16_t>& aa = g->getAAData();
            for (uint32_t z = 0; z < aa.size(); z += 2) {
                if (aa[z+1] + newTop >= 0 && aa[z+1] + newTop < newHeight)
                    ng->addAAHint(aa[z], aa[z+1] + newTop);
            }
            nf->addNewGlyph(i, ng.release());
        }
    }
    return nf;
}
