/**
  *  \file reshack/select.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/select.h"
#include "reshack/painter.h"
#include "util/misc.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "ui/skin.h"

/***************************** RHModeSelector ****************************/

RHModeSelector::RHModeSelector()
{ }

RHModeSelector::~RHModeSelector()
{ }

void
RHModeSelector::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());

    for (int i = 0, e = getNumSlots(); i < e; ++i) {
        GfxRect area(extent.x, extent.y + i * font_heights[FONT_NORMAL],
                     extent.w, font_heights[FONT_NORMAL]);
        if (isActive(i)) {
            drawSolidBar(ctx, area, UIColor::tc_Static);
            ctx.setColor(UIColor::tc_InvStatic);
        } else {
            drawBackground(ctx, area);
            if (isUsable(i))
                ctx.setColor(UIColor::tc_Static);
            else
                ctx.setColor(UIColor::tc_Faded);
        }
        outTextF(ctx, area.x, area.y, area.w, getName(i));
    }

    can.update(extent);
}

bool
RHModeSelector::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Mouse && event.mouse.b != 0 && isInside(GfxPoint(event.mouse.x, event.mouse.y))) {
        setSelectedWidget();
        int n = (event.mouse.y - extent.y) / font_heights[FONT_NORMAL];
        if (n >= 0 && n < getNumSlots()) {
            if (!isActive(n) && isUsable(n)) {
                activate(n);
                drawWidget();
            }
        }
        return true;
    }
    if (event.type == ev_Key) {
        int n = getSlotFromKey(event.key.code);
        if (n >= 0 && !isActive(n) && isUsable(n)) {
            setSelectedWidget();
            activate(n);
            drawWidget();
            return true;
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}

void
RHModeSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(7, getNumSlots());
}

/***************************** RHToolSelector ****************************/

RHToolSelector::RHToolSelector(RHPainter& painter)
    : tools(),
      painter(painter)
{ }
    
RHToolSelector::~RHToolSelector()
{ }

void
RHToolSelector::addNewTool(UIKey key, RHTool* t)
{
    tools.push_back(t);
    keys.push_back(key);
    if (t->isUsable() && !painter.getTool())
        painter.setTool(t);
}

bool
RHToolSelector::isActive(int slot)
{
    return painter.getTool() == tools[slot];
}

bool
RHToolSelector::isUsable(int slot)
{
    return tools[slot]->isUsable();
}

void
RHToolSelector::activate(int slot)
{
    painter.setTool(tools[slot]);
}

string_t
RHToolSelector::getName(int slot)
{
    return tools[slot]->getName();
}

int
RHToolSelector::getNumSlots()
{
    return tools.size();
}

int
RHToolSelector::getSlotFromKey(UIKey k)
{
    for (int i = 0; i < int(keys.size()); ++i)
        if (k == keys[i])
            return i;
    return -1;
}


/***************************** RHZoomSelector ****************************/

static const char*const zoom_names[] = { "1x", "2x", "4x", "8x" };

RHZoomSelector::RHZoomSelector(RHPainter& painter, int nslots)
    : painter(painter), nslots(nslots)
{ }

RHZoomSelector::~RHZoomSelector()
{ }

bool
RHZoomSelector::isActive(int slot)
{
    return painter.getZoom() == slot;
}

bool
RHZoomSelector::isUsable(int /*slot*/)
{
    return true;
}

void
RHZoomSelector::activate(int slot)
{
    painter.setZoom(slot);
}

string_t
RHZoomSelector::getName(int slot)
{
    return zoom_names[slot];
}

int
RHZoomSelector::getNumSlots()
{
    return nslots;
}

int
RHZoomSelector::getSlotFromKey(UIKey k)
{
    if (k >= UIKey('0') && k < UIKey('0'+nslots))
        return k - '0';
    return -1;
}
